from json import load
from os import rename as mv
from ..DIR.DIR import DIR


class File:
    def __init__(self, file_path, json_file_path):
        self.file_path = file_path
        self.__file_name = self.file_path.split('/')[-1]
        self.__file_extension = self.file_path.split('.')[-1]
        self.__json_file = json_file_path
        self.__destination = ''

    def main(self):
        self.__get_file_destination()

        if self.__destination != '':
            self.__move_file_to_destination()

    def __get_directory_to_extensions_dict(self):
        """Gets the directory --> extensions dictionary from the json file"""
        with open(self.__json_file, 'r') as extensions:
            dir_ext = load(extensions)

        return dir_ext

    def __get_file_destination(self):
        """Finds the directory to where the file would go."""
        dir_ext = self.__get_directory_to_extensions_dict()

        for directory, extensions in dir_ext.items():
            if self.__file_extension in extensions:
                self.__destination = directory
                break

    def __move_file_to_destination(self):
        """moves the given file to the destination."""
        try:
            mv(self.file_path, f'{self.__destination}/{self.__file_name}')
        except FileNotFoundError:
            directory = DIR(self.__destination)
            directory.create_directory()
            mv(self.file_path, f'{self.__destination}/{self.__file_name}')



from collections import defaultdict
from json import load, dump
from ..DIR.DIR import DIR


class Extensions:
    def __init__(self, json_file_path, directory_for_extensions=None, extensions=None):
        """
        Loads the extensions file.
        If "directory_for_extensions" and "extensions" were specified, we save them to the json file.
        """

        self._extensions_file_path = json_file_path
        self.extensions = {}
        self.__get_existing_extensions()

        if directory_for_extensions and extensions:
            extensions = set(extensions)
            directory_for_extensions = DIR(directory_for_extensions)

            self.add_extensions(directory_for_extensions, extensions)
            self.write_json_file()

    def __get_existing_extensions(self):
        """Load extensions from json file to the extensions_file dictionary"""
        try:
            with open(self._extensions_file_path, 'r') as extensions:
                x = load(extensions)
                self.extensions = {directory: set(extension) for directory, extension in x.items()}
        except FileNotFoundError:
            pass

    def add_extensions(self, directory: DIR, extensions: set):
        """Adds the extensions to the json object without writing to the file"""
        self.extensions = defaultdict(set, self.extensions)

        for dir_key in self.extensions:
            if extensions <= self.extensions[dir_key]:
                self.extensions[dir_key] -= extensions

        self.extensions[directory.dir_path] |= extensions

    def remove_extensions(self, directory: DIR, extensions: set):
        extensions = set(extensions)
        self.extensions[directory.dir_path] -= extensions
        if not len(self.extensions[directory.dir_path]):
            del self.extensions[directory.dir_path]

    def remove_directory(self, directory: DIR):
        if self.extensions.get(directory.dir_path):
            del self.extensions[directory.dir_path]

    def write_json_file(self):
        """Writes the json object to the json __file_path"""

        # turns all the extension sets into lists since json doesn't work with sets
        extensions_list = {directory: list(extensions) for directory, extensions in self.extensions.items()}

        with open(self._extensions_file_path, 'w') as extensions:
            dump(extensions_list, extensions)



from os import mkdir
from os.path import abspath


class DIR:
    def __init__(self, directory_path):
        self.dir_path = abspath(directory_path)
        self.dir_name = self.dir_path.split('/')[-1]
        self.dir_parent = '/'.join(self.dir_path.split('/')[:-1])

    @property
    def dir_ancestors(self):
        directory = self.dir_parent.split('/')
        ancestors = {'/'.join(directory[:n + 4:]) for n in range(len(directory) - 3)}

        return ancestors

    def create_directory(self):
        try:
            mkdir(self.dir_path)
        except FileExistsError:
            pass



from os import listdir, rmdir
from .DIR import DIR


def test_directory_creation():
    remove_existing_directory('abc')

    new_directory = DIR('abc')
    new_directory.create_directory()

    assert 'abc' in listdir('.')
    remove_existing_directory('abc')


def test_dir_ancestor():
    remove_existing_directory('abc')
    new_directory = DIR('abc')

    assert new_directory.dir_ancestors == {'/HOME/j0e/Projects/Python/automation_scripts/arrange/classes',
                                           '/HOME/j0e/Projects/Python/automation_scripts/arrange',
                                           '/HOME/j0e/Projects/Python/automation_scripts',
                                           '/HOME/j0e/Projects/Python',
                                           '/HOME/j0e/Projects'}


def remove_existing_directory(directory):
    if directory in listdir('.'):
        rmdir(directory)




from ..DIR.DIR import DIR
from .Extensions import Extensions
from os import listdir, remove
from json import load


def test_json_file_creation():
    remove_json_file()

    Extensions('./arrange.json', 'abc', ['abc', 'cba', 'bac'])

    assert 'arrange.json' in listdir('.')
    remove_json_file()


def test_extension_addition():
    remove_json_file()

    ext = Extensions('./arrange.json', 'abc', ['abc', 'cba'])
    ext.add_extensions(DIR('abc'), {'bbc', 'ccb', 'bca'})
    ext.write_json_file()

    with open('arrange.json') as arrange:
        extensions = load(arrange)
        extensions = {dir_: set(ext) for dir_, ext in extensions.items()}

        assert extensions == {DIR('abc').dir_path: {'abc', 'cba', 'bbc', 'ccb', 'bca'}}
    remove_json_file()


def test_extension_removal():
    remove_json_file()

    ext = Extensions('./arrange.json', 'abc', ['abc', 'cba'])
    ext.remove_extensions(DIR('abc'), {'cba'})
    ext.write_json_file()

    with open('arrange.json') as arrange:
        extensions = load(arrange)
        assert extensions == {DIR('abc').dir_path: ['abc']}
    remove_json_file()


def test_same_ext_in_dir():
    Extensions('./arrange.json', 'abc', ['abc', 'abc', 'abc'])

    with open('arrange.json') as arrange:
        extensions = load(arrange)
        assert extensions == {DIR('abc').dir_path: ['abc']}
    remove_json_file()


def test_dir_removal():
    ext = Extensions('./arrange.json', 'abc', ['abc', 'asdf'])
    ext.remove_directory(DIR('abc'))
    ext.write_json_file()

    with open('arrange.json') as arrange:
        json_file = load(arrange)
        assert json_file == {}
    remove_json_file()


def test_extension_duplication_in_child_dirs():
    remove_json_file()
    ext = Extensions('./arrange.json', 'abc', {'alpha', 'beta', 'gama'})
    ext.add_extensions(DIR('abc/b'), {'beta'})
    ext.write_json_file()

    with open('./arrange.json') as json_file:
        x = load(json_file)
        assert x[DIR('abc/b').dir_path] == ['beta'] \
               and \
               set(x[DIR('abc').dir_path]) == {'alpha', 'gama'}
    remove_json_file()


def test_removing_the_only_extension():
    """testing the removal of the only extension in an extension list."""
    remove_json_file()
    ext = Extensions('./arrange.json', 'abc', {'abc'})
    ext.remove_extensions(DIR('abc'), {'abc'})
    ext.write_json_file()

    with open('./arrange.json') as json_file:
        file = load(json_file)
        assert file == {}
    remove_json_file()


def test_extension_duplication_in_separate_dirs():
    """testing the addition of an already-existing extension in a separate directory
    ie: not a child to the directory that already has the extension"""
    remove_json_file()
    abc = DIR('abc').dir_path
    _def = DIR('def')
    ext = Extensions('./arrange.json', 'abc', {'a', 'b'})
    ext.add_extensions(_def, {'a'})
    ext.write_json_file()

    with open('./arrange.json') as json_file:
        file = load(json_file)
        assert file[abc] == ['b'] and 'a' not in file[abc]
        assert file[_def.dir_path] == ['a']

def remove_json_file():
    if 'arrange.json' in listdir('.'):
        remove('arrange.json')



from setuptools import setup, find_packages


setup(
        name='arranger',
        version='1.1.0',
        description="moves each file to its appropriate directory based on the file's extension.",
        author='j0eTheRipper',
        author_email='j0eTheRipper0010@gmail.com',
        url='https://github.com/j0eTheRipper/arranger',
        scripts=['bin/arrange'],
        packages=['engine', 'engine.Extensions', 'engine.File', 'engine.DIR'],
        package_dir={'engine': 'src/engine'},
)
